import random
import functools as fcn


def get_every_city_mutation(mutation_probability):
    '''
    Exchange Every City - each with given probability
    '''

    return fcn.partial(__every_city_mutation, mutation_probability)


def __binomial(n, p):
    ''' maybe innefective binomial distribution '''
    return sum(random.random() < p for _ in range(n))


def __every_city_mutation(mutation_probability, genotype):

    num_genes = len(genotype)
    new_genotype = list(genotype)

    num_exchanged = __binomial(num_genes, mutation_probability)

    idxs = random.sample(range(num_genes), num_exchanged)

    for i, j in zip(idxs, sorted(idxs)):
        new_genotype[i], new_genotype[j] = new_genotype[j], new_genotype[i]

    return tuple(new_genotype)



import pyislands.permutation.tsp.graph as tsp
from pyislands.example.maintsp import solve_tsp_classic, solve_tsp_islands

from math import sqrt


def test_tsp_classic():
    num_cities = 10
    num_iterations = 10
    cities = tsp.random_cities(num_cities)
    adjacency_matrix = tsp.generate_graph(cities)

    solution, penalty = solve_tsp_classic(adjacency_matrix, num_cities, 
                                          num_iterations)

    assert sorted(solution) == [0] + list(range(num_cities))
    assert penalty <= num_cities * sqrt(2)


def test_tsp_islands():
    num_cities = 10
    num_iterations = 10
    cities = tsp.random_cities(num_cities)
    adjacency_matrix = tsp.generate_graph(cities)

    solution, penalty = solve_tsp_islands(adjacency_matrix, num_cities,
                                          num_iterations)

    assert sorted(solution) == [0] + list(range(num_cities))
    assert penalty <= num_cities * sqrt(2)



import random
from itertools import islice, chain


def pmx_find(gen, swath1_dict):
    while gen in swath1_dict:
        gen = swath1_dict[gen]
    return gen


def gen_pmx_child(parent1, swath1, swath2, break1, break2, num_genes):

    swath1_dict = dict(zip(swath1, swath2))

    part1 = (pmx_find(gen, swath1_dict)
             for gen in islice(parent1, break1))
    part2 = (pmx_find(gen, swath1_dict)
             for gen in islice(parent1, break2, num_genes))

    return tuple(chain(part1, swath1, part2))


def partially_mapped_crossover(parent1, parent2):
    '''
    Partially Mapped Crossover
    '''
    num_genes = len(parent1)

    break1 = random.randrange(num_genes)
    break2 = (break1 + random.randrange(num_genes)) % num_genes
    if break1 > break2:
        break1, break2 = break2, break1

    swath1, swath2 = parent2[break1:break2], parent1[break1:break2]

    child1 = gen_pmx_child(parent1, swath1, swath2, break1, break2, num_genes)
    child2 = gen_pmx_child(parent2, swath2, swath1, break1, break2, num_genes)

    return child1, child2



from pyislands.example.evolvetsp import generate_tsp_evolution

from pyislands.types import Island, create_single_island
from pyislands.evolution import finite_evolution, stagnation_evolution

from pyislands import archipelago
from pyislands.archipelago import topology
from pyislands.archipelago import form_destinations
from pyislands.archipelago import assimilation
from pyislands.archipelago.assimilation import get_assimilation
from pyislands.archipelago import migration
from pyislands.archipelago.migration import get_emmigration

import pyislands.permutation.tsp.graph as tsp

from itertools import chain


def solve_tsp_classic(adjacency_matrix, max_stagnation=100, population_size=10,
                      selection_pressure=0.5, mutation_probability=0.6,
                      verbose=True):
    '''
    This functions runs a simple genetic algorithm which evolves a solution of
    a travelling salesman problem. It shows how to print out simple diagnostic
    for population.
    '''

# Simple Genetic Algorithm
    create_population, evolve = \
        generate_tsp_evolution(population_size, selection_pressure,
                               mutation_probability, adjacency_matrix)

    island = create_single_island(create_population, evolve)
    enumerated_evolution = enumerate(stagnation_evolution(max_stagnation, island))
    for iteration, population in enumerated_evolution:
        if verbose:
            best = min(population)
            print("iter = {0}, penalty = {1}".format(iteration, best.penalty))

    return tuple(tsp.form_cycle(best.genotype)), best.penalty


def solve_tsp_islands(adjacency_matrix, num_iterations=10000,
                      use_multiprocess=False, population_size=10,
                      selection_pressure=0.5, mutation_probability=0.6,
                      num_islands=4, migration_size=1, migration_interval=1,
                      degree=1, verbose=True):
    '''
    This functions runs an island genetic algorithm which evolves a solution of
    a travelling salesman problem. It shows how to print out simple diagnostic
    for population.
    '''

# Simple Genetic Algorithm
    create_population, evolve = \
        generate_tsp_evolution(population_size, selection_pressure,
                               mutation_probability, adjacency_matrix)

# Migration Setup
    airports = \
        archipelago.multiprocess.create_airports(num_islands) \
        if use_multiprocess else \
        archipelago.nonparallel.create_airports(num_islands)


    migration_graph = topology.generate_regular(num_islands, degree)  # Ring topology

    destinations = form_destinations(migration_graph, airports)

    assimilations = \
        tuple(get_assimilation(airport, assimilation.policy_2tournament)
              for airport in airports)

    emmigrations = \
        tuple(get_emmigration(migration.random_policy, migration_size, island_destinations)
              for island_destinations in destinations)

    islands = tuple(Island(create_population, evolve,
                           assimilate, emmigrate, migration_interval)
                    for assimilate, emmigrate in zip(assimilations, emmigrations))

    penalty, solution = \
        archipelago.multiprocess.get_solution(islands, num_iterations, verbose) \
        if use_multiprocess else \
        archipelago.nonparallel.get_solution(islands, num_iterations, verbose)

    return tuple(tsp.form_cycle(solution)), penalty


def main_tsp(num_cities=100, num_iterations=100, use_islands=False,
             use_multiprocess=False, verbose=True):
    '''
    This functions creates a random travelling salesman problem and
    solves it using classical or island model genetic algorithm.
    '''
    cities = tsp.random_cities(num_cities)
    adjacency_matrix = tsp.generate_graph(cities)

    if use_islands:
        solution, penalty = solve_tsp_islands(adjacency_matrix, num_iterations,
                                              use_multiprocess=use_multiprocess,
                                              verbose=verbose)
    else:
        solution, penalty = solve_tsp_classic(adjacency_matrix)

    print("solution = {0}".format(solution))
    print("penalty = {0}".format(penalty))


if __name__ == '__main__':
    main_tsp(num_cities=100, use_islands=True, use_multiprocess=False)



'''
Module containing selection functions on population
'''
import random
from math import floor, sqrt

import functools as fcn


def ktournament(k, population):
    '''
    randomly select k-individuals from the population
    and sort them by penalty in the increasing order
    '''

    enumerated_population = list(enumerate(population))

# enumerated_individual = (idx, (penalty, genotype))
    selected_individuals = random.sample(enumerated_population, k)

    penalty = lambda enumerated_individual: enumerated_individual[1][0]

# individuals are sorted by penalty
    idxs, individuals = zip(*sorted(selected_individuals, key=penalty))

    return individuals, idxs


def __ranking_select(sp, population):
    ''' selects random individual with distribution (sp in [1, 2]) '''

    n = len(population)
    p = random.random()

    a = sp - 1
    b = (2 - sp) * n - 1
    c = -n * (n - 1) * p

    idx = (-b + sqrt(b * b - 4 * a * c)) / (2 * a)
    idx = int(floor(idx))

    individual = population[idx]

    return individual


def get_ranking_select(selection_pressure, population):
    '''
    It sorts the population and returns a closure around __ranking_select
    function.  This closure returns a random individual from the population
    with distriburion based on penalty determined with selection_pressure.

    selection_pressure in [0, 1]:
        0 -> no selection pressure (each individual can be selected with equal
        probability)
        1 -> maximum selection pressure (the worst individual cannot be chosen)

    The probability of chosing index i in (0, 1, ..., n - 1) in the sorted
    population (sp in [1, 2]):
        p_i = (2 - sp  + 2 * (sp - 1) * i / (n - 1)) / n

        Marko Cupic, Prirodom inspirirani optimizacijski algoritmi
    '''

    sp = selection_pressure + 1

# the best individual has the largest index
    population = sorted(population, reverse=True)

    return fcn.partial(__ranking_select, sp, population)



'''
This module contains examples of solving a random travelling salesman
problem using pyislands.
'''
from pyislands import ga
from pyislands.creation import get_create_population

from pyislands.selection import get_ranking_select
from pyislands.permutation.generate import get_number_permutation_generator
from pyislands.permutation.mutation.rsm import get_reversed_sequence_mutation
from pyislands.permutation.crossover.ox1 import order_crossover1
import pyislands.permutation.tsp.graph as tsp

import functools as fcn


def generate_tsp_evolution(population_size, selection_pressure,
                           mutation_probability, adjacency_matrix):
    '''
    This function defines algorithm used for solving a travelling salesman
    problem. Algorithm parameters are hardcoded into this function, it is just
    an example.
    '''

    evaluate = tsp.get_evaluate_cycle(adjacency_matrix)

    generate = get_number_permutation_generator(1, len(adjacency_matrix))

    mutate = get_reversed_sequence_mutation(mutation_probability)

    crossover = order_crossover1

    get_select = fcn.partial(get_ranking_select, selection_pressure)
    evolve = ga.generational.get_elitist_evolution(get_select, crossover, mutate, evaluate)

    create_population = get_create_population(generate, evaluate, population_size)

    return create_population, evolve



'''
Module containing an implementation of generational genetic algorithm.
'''
from pyislands.types import create_individual

import functools as fcn
from  itertools import chain


def get_elitist_evolution(get_select, crossover, mutate, evaluate):
    '''
    Returns closure evolve:
        evolve - uses get_select, crossover, mutate and evaluate
                 to evolve some population
               - generational iteration

    population_k = evolve(population_k-1)
    '''

    return fcn.partial(__elitist_evolve, get_select, crossover, mutate, evaluate)


def __create_child(select, crossover, mutate, evaluate, dummy):
    ''' create one child '''

    mother, father = select(), select()

# Crossover
    children_genotypes = crossover(mother.genotype, father.genotype)
    best_genotype = min(children_genotypes, key=evaluate)

# Mutation
    new_genotype = mutate(best_genotype)

    child = create_individual(evaluate, new_genotype)

    return child


def __elitist_evolve(get_select, crossover, mutate, evaluate, population):
    '''
    This function uses crossover, mutate and evalute, functions
    passed as arguments to get_steady_evolve.

    Preserves one elite individual.
    '''

    select = get_select(population)

    create_child = fcn.partial(__create_child, select, crossover,
                               mutate, evaluate)

    children_population = map(create_child, population[:-1])

    best_individual = min(population)

    new_population = tuple(chain(children_population, [best_individual]))

    return new_population



from itertools import islice


def evolution(island):
    '''
    Infinite generator for evolution of some population.
    This generator yields population:
        population - tuple together containing tuples/individuals

    population_0 = create()
    population_1 = evolve(population_0, info_0)
    .
    .
    population_k = evolve(population_k-1, info_k-1)
    .
    .

    Since population is a tuple/an immutable type, a population cannot be
    influenced by outside functions. Population can be used only to gather
    statistics

    If no immigration and emmigration is used this island evolution
    becomes a classical genetic algorithm.
    '''

    population = island.create_population()

    while True:
        for _ in range(island.migration_interval if island.migrate else 1):
            yield population
# Immigration - Outside individuals are inhabiting an island
            if island.assimilate:
                population = island.assimilate(population)

# Evolution - Each island population is evolved into the next generation
            population = island.evolve(population)

# Emmigration - Sends individuals (clones) from one population onto voyage
        if island.migrate:
            island.migrate(population)


def finite_evolution(num_iterations, island):
    ''' Same as evolution, except stopped after num_iterations '''
    return islice(evolution(island), num_iterations)


def stagnation_evolution(max_stagnation, island):
    ''' Same as evolution, except stopped after max_stagnation '''

    infinite_evolution = evolution(island)

    population = next(infinite_evolution)

    best = min(population)

    stagnation = 0

    while stagnation < max_stagnation:
        stagnation += 1
        yield population
        population = next(infinite_evolution)

        current_best = min(population)
        if current_best < best:
            stagnation = 0
            best = current_best



from pyislands.permutation.tsp.graph import generate_graph, get_evaluate_path
from nose.tools import assert_almost_equals
from math import sqrt
import functools as fcn


class TestTSPGraphSimple:

    def __init__(self):
        cities = [(0.0, 0.0), (1.0, 0.0)]
        self.adjacency_matrix = generate_graph(cities)

    def test_distance(self):
        d1 = self.adjacency_matrix[0][1]
        d2 = self.adjacency_matrix[1][0]
        assert_almost_equals(d1, 1.0)
        assert_almost_equals(d2, 1.0)

    def test_evaluate_cycle(self):
        evaluate = get_evaluate_path(self.adjacency_matrix)

        length1 = evaluate([0, 1])
        assert_almost_equals(length1, 1.0)

        length2 = evaluate([0, 1, 0])
        assert_almost_equals(length2, 2.0)

        length3 = evaluate([1, 0])
        assert_almost_equals(length3, 1.0)

        length4 = evaluate([1, 0, 1])
        assert_almost_equals(length4, 2.0)


class TestTSPGraphMedium:
    '''
    test done on three cities
    '''

    def __init__(self):
        cities = [(0.0, 0.0), (1.0, 0.0), (1.0, 1.0)]
        self.adjacency_matrix = generate_graph(cities)

    def test_distance(self):
        test_edges = {(0, 1): 1.0, (1, 2): 1.0, (2, 0): sqrt(2.0)}

        for (city1, city2) in test_edges:
            distance = self.adjacency_matrix[city1][city2]
            assert_almost_equals(distance, test_edges[(city1, city2)])

    def test_evaluate_path(self):
        evaluate = get_evaluate_path(self.adjacency_matrix)

        length = evaluate([0, 1, 2, 0])
        assert_almost_equals(length, 2.0 + sqrt(2.0))

        length = evaluate([2, 1, 0, 2])
        assert_almost_equals(length, 2.0 + sqrt(2.0))



import random
import functools as fcn
from itertools import chain, islice


def get_reversed_sequence_mutation(mutation_probability):
    '''
    Reversed Sequence Mutation - ignored with
    probability (1 - mutation_probability)

    In tsp, equivalent to exchange of two edges.
    '''

    return fcn.partial(__reversed_sequence_mutation, mutation_probability)


def __reversed_sequence_mutation(mutation_probability, genotype):

    if not random.random() < mutation_probability:
        return genotype

    num_genes = len(genotype)
    break1, break2 = sorted(random.sample(range(num_genes + 1), 2))

    new_genotype = tuple(chain(islice(genotype, break1),
                               reversed(genotype[break1:break2]),
                               islice(genotype, break2, num_genes)))

    return new_genotype



'''
Synchronous Island creation for testing on a single computer,
this module connects populations/islands using simple queues/lists.
'''
from pyislands.evolution import finite_evolution

import sys
if sys.version_info[0] == 2:
    import Queue as queue
    from itertools import izip as zip
else:
    import queue
from itertools import islice


def create_airports(num_islands):
    '''
    An airport is a traffic stop for an islands. An airport
    should support get, put and empty operations. Queues
    are used as an implementation of an airport.
    '''

    airports = tuple(queue.Queue() for _ in range(num_islands))

    return airports



def get_solution(islands, num_iterations, verbose=True):
    '''
    Utility function used for getting a solutions from a single process algorithm.
    '''


    evolutions = (finite_evolution(num_iterations, island)
                  for island in islands)

    for iteration, populations in enumerate(zip(*evolutions)):
        if verbose:
            best_individuals = map(min, populations)
            least_penalty, _ = min(best_individuals)
            print("iteration = {0}, penalty = {1}".
                  format(iteration, least_penalty))

    return min(map(min, populations))



from math import sqrt
from random import random
import itertools
from itertools import chain
import functools as fcn


def generate_graph(cities):
    '''
    Generates Travelling Salesman's Graph (undirected graph)
    given cities salesman needs to visit

    cities  - [(x1, y1), (x2, y2), ..., (xn, yn)]
            - locations of n cities
            - n = len(cities)
    '''
    adjacency_matrix = \
        [[sqrt((x2-x1) ** 2 + (y2 - y1) ** 2) for (x1, y1) in cities]
         for (x2, y2) in cities]

    return adjacency_matrix


def __evaluate_path(adjacency_matrix, tsp_path):
    '''
    function for evaluation of cycle given list

    tsp_path = [city1, city2, ..., cityn]
    '''
    iter0, iter1 = itertools.tee(tsp_path)
    next(iter1, None)
    return sum(adjacency_matrix[city1][city2]
               for city1, city2 in zip(iter0, iter1))


def get_evaluate_path(adjacency_matrix):
    ''' evaluate tsp path '''
    return fcn.partial(__evaluate_path, adjacency_matrix)


def form_cycle(tsp_path, start_city=0):
    '''
    returns an iterator containing cycle
    [start_city] -> tsp_path -> [start_city]
    '''
    start_list = [start_city]
    return chain(start_list, tsp_path, start_list)


def __evaluate_cycle(adjacency_matrix, tsp_path):
    ''' evaluate path that is part of the cycle '''
    nul = [0]
    return __evaluate_path(adjacency_matrix, form_cycle(tsp_path))


def get_evaluate_cycle(adjacency_matrix):
    ''' evaluate path that is part of the cycle '''
    return fcn.partial(__evaluate_cycle, adjacency_matrix)


def random_cities(num_cities=100):
    '''
    create graph for Travelling Salesman
    Problem with n cities randomly distributed on
    square [0, 1) x [0, 1)
    '''
    x = [random() for _ in range(num_cities)]
    y = [random() for _ in range(num_cities)]
    cities = list(zip(x, y))

    return cities



from pyislands.evolution import finite_evolution

import multiprocessing
import functools as fcn


def create_airports(num_islands):
    '''
    An airport is a traffic stop for an islands. An airport
    should support get, put and empty operations. Queues
    are used as an implementation of an airport.
    '''

    manager = multiprocessing.Manager()
    airports = tuple(manager.Queue() for _ in range(num_islands))

    return airports


def __get_solution(num_iterations, island, verbose=False):
    for population in finite_evolution(num_iterations, island):
        if verbose:
            best = min(population)
            print("penalty = {0}".format(best.penalty))
    return min(population)


def get_solution(islands, num_iterations, verbose):
    '''
    Utility function used for getting a solutions from a single process algorithm.
    '''

    num_islands = len(islands)
    pool = multiprocessing.Pool(processes=num_islands)

    solution_after_niter = fcn.partial(__get_solution, num_iterations,
                                       verbose=verbose) 

    solutions = pool.map(solution_after_niter, islands)

    return min(solutions)



'''
Module contains functions for generationg islands topology.
'''

def generate_regular(num_islands, degree):
    '''
    Creates a directed, regular, fully connected graph with
    given degree of each vertex/island. This function is used
    for creating migration topology, return its adjacency list.
    '''
    idxs = list(range(num_islands))
    neighbors = [idxs[-i:] + idxs[:-i] for i in range(degree, 0, -1)]

    return list(zip(*neighbors))



from pyislands.permutation.crossover.pmx import partially_mapped_crossover
from pyislands.permutation.crossover.pmx import pmx_find
from pyislands.permutation.crossover.pmx import gen_pmx_child


def test_pmx():
    parent1_genotype = (0, 1, 2, 3, 4, 5, 6)
    parent2_genotype = (6, 4, 0, 2, 1, 5, 3)

    children_genotypes = partially_mapped_crossover(parent1_genotype,
                                                    parent2_genotype)

    for genotype in children_genotypes:
        for i in parent1_genotype:
            assert i in genotype


def test_pmx_find():
    swath1 = (0, 1, 2, 3)
    swath2 = (1, 2, 3, 4)

    swath1_dict = dict(zip(swath1, swath2))
    swath2_dict = dict(zip(swath2, swath1))

    assert pmx_find(0, swath1_dict) == 4
    assert pmx_find(2, swath2_dict) == 0


def test_pmx_child():
    parent1 = (0, 1, 2, 3, 4, 5, 6)
    parent2 = (6, 4, 0, 2, 1, 5, 3)
    break1, break2 = 1, 6

    swath1, swath2 = parent2[break1:break2], parent1[break1:break2]

    child1 = gen_pmx_child(parent1, swath1, swath2, break1, break2,
                           len(parent1))

    assert child1 == (3, 4, 0, 2, 1, 5, 6)



import random
import functools as fcn


def get_number_permutation_generator(start, end):
    ''' Closure around elements (start, start + 1, ..., end - 1) '''
    return get_random_permutation_generator(range(start, end))


def get_random_permutation_generator(elements):
    '''
    Generates closure around elements = (x_0, x_1, x_2, ..., x_n-1). This
    closure returns random permutation of this elements,
    e.g. (x_0, x_n-1, x_3, ..., x_2, x_n - 1)
    '''

    generator_elements = list(elements)

    return fcn.partial(__random_permutation_generator, generator_elements)


def __random_permutation_generator(generator_elements):
    ''' Closure which returns permutation of generator_elements '''

    random.shuffle(generator_elements)

    return tuple(generator_elements)


def unit(num_elements):
    ''' generates unit permutation, (0, 1, 2, ..., n - 1) '''
    return tuple(range(num_elements))



from collections import namedtuple


Individual = namedtuple('Individual', ['penalty', 'genotype'])

def create_individual(evaluate, genotype):
    ''' Utility function used for evaluating genotype '''

    penalty = evaluate(genotype)

    return Individual(penalty, genotype)


Island = namedtuple('Island', ['create_population', 'evolve',
                               'assimilate', 'migrate', 'migration_interval'])

def create_single_island(create_population, evolve):
    return Island(create_population, evolve, None, None, None)



from pyislands.types import create_individual

import functools as fcn


def __create_population(generate, evaluate, population_size):
    genotypes = tuple(generate() for _ in range(population_size))
    population = tuple(map(fcn.partial(create_individual, evaluate), genotypes))

    return population


def get_create_population(generate, evaluate, population_size):
    '''
    This function is used in evolution to create an initial
    population.
    '''

    return fcn.partial(__create_population, generate, evaluate, population_size)



from pyislands.permutation.tsp.mutation import get_twoopt_mutation
from pyislands.permutation.tsp.graph import generate_graph, random_cities
from pyislands.permutation.tsp.graph import get_evaluate_cycle

def create_random_twoopt():
    adjacency_matrix = generate_graph(random_cities(10))

    two_opt = get_twoopt_mutation(adjacency_matrix, 1.0)

    return two_opt, adjacency_matrix


def test1_two_opt():
    two_opt, _ = create_random_twoopt()
    genotype = range(1, 10)
    new_genotype = two_opt(genotype)

    for i in genotype:
        assert i in new_genotype


def test1_two_opt():
    two_opt, adjacency_matrix = create_random_twoopt()
    genotype = range(1, 10)

    new_genotype = two_opt(genotype)

    evaluate = get_evaluate_cycle(adjacency_matrix)

    assert evaluate(new_genotype) < evaluate(genotype)



import random
import functools as fcn
from itertools import islice, chain, combinations


def get_twoopt_mutation(adjacency_matrix, mutation_probability):
    return fcn.partial(__twoopt_mutation, adjacency_matrix,
                       mutation_probability)


def __twoopt_mutation(adjacency_matrix, mutation_probility, genotype):
    '''
    After each valid exchange_cost(i, j), tsp_cycle:

    city_0, city_1, city_2, ..., city_i-1, city_i, ... city_j-1, city_j, ...,
    city_n-1, city_0

    becomes

    city_0, city_1, city_2, ..., city_i-1, city_j, city_j-1, ..., city_i,
    city_j+1..., city_n-1, city_0
    '''

    if not random.random() < mutation_probility:
        return genotype

    tsp_cycle = list(chain([0], genotype, [0]))
    n = len(tsp_cycle)

    distance = lambda idx1, idx2: adjacency_matrix[tsp_cycle[idx1]][tsp_cycle[idx2]]

    def exchange_cost(i, j):

        removed_cost = distance(i - 1, i) + distance(j - 1, j)
        added_cost = distance(i - 1, j - 1) + distance(i, j)

        return added_cost - removed_cost

    exchange = True
    idxs = range(1, n - 1)
    while exchange:
        exchange = False
        for i, j in combinations(idxs, 2):
            if j > i + 1 and exchange_cost(i, j) < 0:

                tsp_cycle = list(chain(islice(tsp_cycle, 0, i),
                                       reversed(tsp_cycle[i:j]),
                                       islice(tsp_cycle, j, n)))
                exchange = True

    new_genotype = tuple(islice(tsp_cycle, 1, n - 1))

    return new_genotype



__author__ = 'Slaven Glumac'
__email__ = 'slaven.glumac@gmail.com'
__version__ = '0.1.2'

import pyislands.types
import pyislands.ga
import pyislands.evolution
import pyislands.creation
import pyislands.archipelago

import pyislands.permutation
import pyislands.selection

import pyislands.example



from pyislands.permutation.generate import get_random_permutation_generator
from pyislands.permutation.generate import unit


def test_random_permutation():
    ''' This function asserts containing elements in permutation mutation '''
    elements = list(range(5, 20))
    generate = get_random_permutation_generator(elements)

    genotype = generate()

    for i in elements:
        assert i in genotype


def test_unit():
    ''' This function asserts containing elements in permutation mutation '''

    for i in unit(10):
        assert i in range(10)



import random
from itertools import chain, islice


def cyclic_crossover(parent1, parent2):
    '''
    Cyclic Crossover
    '''
    num_genes = len(parent1)

    child1 = [-1] * num_genes
    child2 = [-1] * num_genes

    while -1 in child1:
        i = child1.index(-1)
        while parent2[i] not in child1:
            j = parent1.index(parent2[i])
            child1[j] = parent1[j]
            child2[j] = parent2[j]
            i = j
        parent1, parent2 = parent2, parent1

    return tuple(child1), tuple(child2)



from itertools import izip


def geom(x0, q, xmax):
    x = x0
    while x < xmax:
        yield x
        x *= q

def bottom_up_merge(xs, ys, ileft, iright, iend):
    i0 = ileft
    i1 = iright
    num_inversions = 0
    for j in xrange(ileft, iend):
        if i0 < iright and (i1 >= iend or xs[i0] <= xs[i1]):
            ys[j] = xs[i0]
            i0 += 1
        else:
            ys[j] = xs[i1]
            i1 += 1
            num_inversions += iright - i0
    return num_inversions


def bottom_up_sort(xs):
    n = len(xs)
    ys = list(xs)
    tmp = range(n)
    inv = 0
    for width in geom(1, 2, n):
        for ileft in xrange(0, n, 2 * width):
            iright = min(ileft + width, n)
            iend = min(ileft + 2 * width, n)
            inv += bottom_up_merge(ys, tmp, ileft, iright, iend)

        ys, tmp = tmp, ys

    return ys, inv


def inversions_on2(xs):
    n = len(xs)
    inv = 0
    for i in range(n):
        for j in range(i + 1, n):
            inv += xs[i] > xs[j]
    return inv

def distance(px, py):
    '''
    distance between two permutations px and py,
    measured as a number of inversions of px * py^-1
    '''

# inverse - ipy = py^-1
    #import ipdb; ipdb.set_trace() # BREAKPOINT
    _, ipy = izip(*sorted((e, i) for i, e in enumerate(py)))
# apply px to inverse - pr = px * ipy
    pr = [px[i] for i in ipy]

    _, inversions = bottom_up_sort(pr)

    return inversions


def main():
    n = 1000
    xs = range(n)
    import random
    random.shuffle(xs)
    print xs
    inv = inversions_on2(xs)
    print "real inv =", inv
    ys, inv = bottom_up_sort(xs)
    print ys
    print "xs is sorted =", ys == range(n)
    print "inversions =", inv


if __name__ == '__main__':
    main()



import random
from itertools import chain, islice


def order_merge(swath, parent2, break1, break2, num_genes):
    ''' merge order '''
    swath_set = set(swath)
    in_swath = lambda gene: gene not in swath_set

    inv_parent2 = chain(islice(parent2, break2, num_genes),
                        islice(parent2, break2))

    other_genes = list(filter(in_swath, inv_parent2))

    child = tuple(chain(islice(other_genes, num_genes - break2,
                               num_genes - break2 + break1),
                        swath,
                        islice(other_genes, num_genes - break2)))
    return child


def order_crossover1(parent1, parent2):
    '''
    Cyclic Crossover
    '''
    num_genes = len(parent1)

    break1 = random.randrange(num_genes)
    break2 = (break1 + random.randrange(num_genes)) % num_genes
    if break1 > break2:
        break1, break2 = break2, break1

    swath1, swath2 = parent2[break1:break2], parent1[break1:break2]

    child1 = order_merge(swath1, parent2, break1, break2, num_genes)
    child2 = order_merge(swath2, parent1, break1, break2, num_genes)

    return child1, child2



'''
Package contains implementations of archipelago, network of islands.
Archipelago may be spread accross multiple processes, multiple computers,
but may be implemented on a single process computer.

Islands in archipelago is presented with its integer id.
'''

def form_destinations(migration_graph, airports):
    '''
    This functions forms destinations for each island emmigration
    based on adjacency matrix migration_graph. It returns tuple of tuples:

    destinations = (destinations_island_0,
                    destinations_islands_1,
                    .
                    .
                    destinations_islands_n)

    destinations_island_0 = (dst_00, dst_01, ..., dst_0m)
    destinations_island_1 = (dst_10, dst_11, ..., dst_0m)
    .
    .
    destinations_island_n = (dst_n0, dst_n1, ..., dst_0m)

    where dst_ij is an airport on island_j
    '''

    destinations = tuple(tuple(airports[idx] for idx in idxs)
                         for idxs in migration_graph)
    
    return destinations


import pyislands.archipelago.assimilation
import pyislands.archipelago.migration

import pyislands.archipelago.topology
import pyislands.archipelago.nonparallel
import pyislands.archipelago.multiprocess



import random
import functools as fcn


def get_emmigration(policy, migration_size, destinations):
    '''
    This function returns emmigrate function:
        emmigrate(population)
    '''
    return fcn.partial(policy, migration_size, destinations)


def random_policy(migration_size, destinations, population):
    '''
    This policy chooses random individuals for migration and
    random destination for each of these individuals.

    After emmigration there are two copies of one individual, one
    in the original population and the other at the destination.
    '''

    emmigrants = random.sample(population, migration_size)

    for individual in emmigrants:
        destination = random.choice(destinations)
        destination.put(individual)



'''
Module contains mutation generators for permutation mutations.
Mutation is function with signature:

    def mutation(genotype):
        ...
        return mutated_genotype

'''
import pyislands.permutation.mutation.ecm
import pyislands.permutation.mutation.rsm



from pyislands.permutation.mutation.rsm import get_reversed_sequence_mutation
from pyislands.permutation.mutation.ecm import get_every_city_mutation

from tests.permutation.mutation import check_mutation


def test_ecm():
    check_mutation(get_every_city_mutation(1.0))



from pyislands.selection import ktournament

import functools as fcn


def get_assimilation(airport, policy):
    '''
    This function returns assimilate function:
        new_population = assimilate(population)
    '''
    return fcn.partial(__assimilate, policy, airport)


def __assimilate(policy, airport, population):
    '''
    Assimilation imports all immigrants from the airport.
    The airport should support get, put and empty methods (e.g. Queue.Queue or
    multiprocessing.Queue).

    Immigrants are then mixed with the population using migration policy:
        new_population = policy(population, immigrants)
    '''

# import immigrants
    immigrants = []
    while not airport.empty():
        genotype = airport.get()
        immigrants.append(genotype)

# mix immigrants with the population
    new_population = policy(population, immigrants)

    return new_population


def policy_2tournament(population, immigrants):
    ''' assimilation policy of some population '''

    new_population = list(population)

    for immigrant in immigrants:
        _, idxs = ktournament(2, population)
        bad_idx = idxs[1]

        new_population[bad_idx] = immigrant

    return tuple(new_population)



from pyislands.permutation.crossover.ox1 import order_merge, order_crossover1


def test_order_merge_ox1():
    parent1 = (1, 2, 3, 4, 5, 6, 7, 8)
    parent2 = (2, 4, 6, 8, 7, 5, 3, 1)

    num_genes = len(parent1)

    break1, break2 = 2, 5

    swath1, swath2 = parent1[break1:break2], parent2[break1:break2]

    child1 = order_merge(swath1, parent2, break1, break2, num_genes)
    child2 = order_merge(swath2, parent1, break1, break2, num_genes)

    assert child1 == (8, 7, 3, 4, 5, 1, 2, 6)
    assert child2 == (4, 5, 6, 8, 7, 1, 2, 3)


def test_pmx():
    parent1_genotype = (0, 1, 2, 3, 4, 5, 6)
    parent2_genotype = (6, 4, 0, 2, 1, 5, 3)

    children_genotypes = order_crossover1(parent1_genotype, parent2_genotype)

    for genotype in children_genotypes:
        for i in parent1_genotype:
            assert i in genotype



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


readme = open('README.rst').read()
history = open('docs/history.rst').read().replace('.. :changelog:', '')

requirements = [
    # no requirements
]

test_requirements = [
    'nose'
]

setup(
    name='pyislands',
    version='0.1.2',
    description='PyIslands contains an implementation of islands genetic algorithms.',
    long_description=readme + '\n\n' + history,
    author='Slaven Glumac',
    author_email='slaven.glumac@gmail.com',
    url='https://github.com/sglumac/pyislands',
    packages=['pyislands', 'pyislands.ga', 'pyislands.archipelago',
              'pyislands.example', 'pyislands.permutation',
              'pyislands.permutation.crossover',
              'pyislands.permutation.mutation',
              'pyislands.permutation.tsp'],
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords='pyislands',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



'''
Module contains permutation crossovers. Crossover is function
with signature:

    def crossover(penalty_function, parent1, parent2):
        ...
        return children

where object children is list of individuals.
'''
import pyislands.permutation.crossover.pmx
import pyislands.permutation.crossover.cx
import pyislands.permutation.crossover.ox1



def check_mutation(mutation):
    ''' This function asserts containing elements in permutation mutation '''
    genotype = range(10)

    new_genotype = mutation(genotype)

    for i in genotype:
        assert i in new_genotype


import tests.permutation.mutation.testrsm
import tests.permutation.mutation.testecm



'''
Module containing an implementation of steady-state genetic algorithm.
'''
from pyislands.selection import ktournament
from pyislands.types import create_individual

import functools as fcn


def get_evolution(crossover, mutate, evaluate):
    '''
    Returns closure evolve:
        evolve - uses crossover, mutate and evaluate to evolve some population
               - steady state iteration, uses 3-tournament selection

    population_k = evolve(population_k-1)
    '''

    return fcn.partial(__evolve, crossover, mutate, evaluate)


def __evolve(crossover, mutate, evaluate, population):
    '''
    This function uses crossover, mutate and evalute, functions
    passed as arguments to get_steady_evolve.
    '''

# Selection = 3-Tournament
    individuals, idxs = ktournament(3, population)
    parent1, parent2 = individuals[0:2]
    bad_idx = idxs[-1]

# Crossover
    children_genotypes = crossover(parent1.genotype, parent2.genotype)
    best_genotype = min(children_genotypes, key=evaluate)

# Mutation
    new_genotype = mutate(best_genotype)
    penalty = evaluate(new_genotype)

    child = create_individual(evaluate, new_genotype)

# Replace bad individual
    new_population = list(population)
    new_population[bad_idx] = child

    return tuple(new_population)
